# -------------------------------------------------------------------------------------------------
#  Copyright (C) 2015-2021 Nautech Systems Pty Ltd. All rights reserved.
#  https://nautechsystems.io
#
#  Licensed under the GNU Lesser General Public License Version 3.0 (the "License");
#  You may not use this file except in compliance with the License.
#  You may obtain a copy of the License at https://www.gnu.org/licenses/lgpl-3.0.en.html
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# -------------------------------------------------------------------------------------------------

import heapq

from nautilus_trader.core.correctness cimport Condition
from nautilus_trader.core.functions cimport bisect_double_right
from nautilus_trader.model.c_enums.depth_type cimport DepthType
from nautilus_trader.model.objects cimport Price
from nautilus_trader.model.objects cimport Quantity
from nautilus_trader.model.orderbook.level cimport Level
from nautilus_trader.model.orderbook.order cimport Order


cdef class Ladder:
    """
    Represents a ladder of orders in a book.
    """
    def __init__(
        self,
        bint reverse,
        int price_precision,
        int size_precision,
    ):
        """
        Initialize a new instance of the ``Ladder`` class.

        Parameters
        ----------
        reverse : bool
            If the ladder should be represented in reverse order of price (bids).
        price_precision : int
            The price precision for the book.
        size_precision : int
            The size precision for the book.

        Raises
        ------
        ValueError
            If price_precision is negative (< 0).
        ValueError
            If size_precision is negative (< 0).

        """
        Condition.not_negative_int(price_precision, "price_precision")
        Condition.not_negative_int(size_precision, "size_precision")

        self._order_id_level_index = {}  # type: dict[str, Level]

        self.levels = []  # type: list[Level]  # TODO: Make levels private??
        self.reverse = reverse
        self.price_precision = price_precision
        self.size_precision = size_precision

    def __repr__(self) -> str:
        return f"{Ladder.__name__}({self.levels})"

    cpdef void add(self, Order order) except *:
        """
        Add the given order to the ladder.

        Parameters
        ----------
        order : Order
            The order to add.

        """
        Condition.not_none(order, "order")

        cdef list existing_prices = self.prices()

        cdef int price_idx
        cdef Level level
        if order.price in existing_prices:
            # Level exists, add new order
            price_idx = existing_prices.index(order.price)
            level = self.levels[price_idx]
            level.add(order=order)
        else:
            # New price, create Level
            level = Level(price=order.price)
            level.add(order)

            if self.reverse:
                self.levels.append(level)
                # TODO: heapq._siftdown_max is temporary before custom data structure
                heapq._siftdown_max(self.levels, 0, len(self.levels) - 1)
            else:
                price_idx = bisect_double_right(existing_prices, level.price)
                self.levels.insert(price_idx, level)

        self._order_id_level_index[order.id] = level

    cpdef void update(self, Order order) except *:
        """
        Update the given order in the ladder.

        Parameters
        ----------
        order : Order
            The order to add.

        """
        Condition.not_none(order, "order")

        if order.id not in self._order_id_level_index:
            self.add(order=order)
            return

        # Find the existing order
        cdef Level level = self._order_id_level_index[order.id]
        if order.price == level.price:
            # This update contains a volume update
            level.update(order=order)
            if not level.orders:
                self.levels.remove(level)
        else:
            # New price for this order, delete and insert
            self.delete(order=order)
            self.add(order=order)

    cpdef void delete(self, Order order) except *:
        """
        Delete the given order in the ladder.

        Parameters
        ----------
        order : Order

        Raises
        ------
        KeyError
            If order.id is not contained in the order_id_level_index.

        """
        Condition.not_none(order, "order")

        cdef Level level = self._order_id_level_index.get(order.id)
        if level is None:
            return
            # TODO: raise KeyError("Cannot delete order: not found at level.")
        cdef int price_idx = self.prices().index(level.price)
        level.delete(order=order)
        self._order_id_level_index.pop(order.id)
        if not level.orders:
            del self.levels[price_idx]

    cpdef list depth(self, int n=1):
        """
        Return the levels in the ladder to the given depth.

        Parameters
        ----------
        n : int
            The maximum level to query.

        Returns
        -------
        list[Level]

        """
        if not self.levels:
            return []
        n = n or len(self.levels)
        return self.levels[:n]

    cpdef list prices(self):
        """
        The prices in the ladder.

        Returns
        -------
        list[double]

        """
        return [level.price for level in self.levels]

    cpdef list volumes(self):
        """
        The volumes in the ladder.

        Returns
        -------
        list[double]

        """
        return [level.volume() for level in self.levels]

    cpdef list exposures(self):
        """
        The exposures in the ladder.

        Returns
        -------
        list[double]

        """
        return [level.exposure() for level in self.levels]

    cpdef Level top(self):
        """
        The top `Level` in the ladder.

        Returns
        -------
        Level or None

        """
        cdef list top = self.depth(1)
        if top:
            return top[0]
        else:
            return None

    cpdef list simulate_order_fills(self, Order order, DepthType depth_type=DepthType.VOLUME):
        """
        Return a simulation of where this order would be filled in the ladder.

        Parameters
        ----------
        order : Order
            The order to simulate.
        depth_type : DepthType
            The depth type to simulate.

        Returns
        -------
        list[(Price, Quantity)]

        """
        Condition.not_none(order, "order")

        cdef list fills = []
        cdef double cumulative_denominator = 0.0
        cdef double current = 0.0
        cdef double target = order.size if depth_type == DepthType.VOLUME else order.price * order.size
        cdef Level level
        cdef Order book_order

        for level in self.levels:
            if self.reverse and level.price < order.price:
                break
            elif not self.reverse and level.price > order.price:
                break
            for book_order in level.orders:
                current = book_order.size if depth_type == DepthType.VOLUME else book_order.exposure()
                if (cumulative_denominator + current) >= target:
                    # This order has filled us, calc and return
                    remainder = target - cumulative_denominator
                    fills.append((
                        Price(book_order.price, precision=self.price_precision),
                        Quantity(remainder, precision=self.size_precision),
                    ))
                    cumulative_denominator += remainder
                    return fills
                else:
                    # Add this order and continue
                    fills.append((
                        Price(book_order.price, precision=self.price_precision),
                        Quantity(current, precision=self.size_precision),
                    ))
                    cumulative_denominator += current

        return fills
